@@ -25,12 +25,8 @@
 #include <ctype.h>
 #include <string.h>
 
-typedef int (*prog)(
-#if NeedFunctionPrototypes
-		    Display* display, int argc, char *argv[],
-		    char *prog_name, char *prog_desc
-#endif
-);
+typedef int (*prog)(Display* display, int argc, char *argv[],
+		    char *prog_name, char *prog_desc);
 
 typedef struct
 {
@@ -106,6 +102,10 @@ static entry drivers[] =
       "<window> <device>",
       set_clientpointer
     },
+    { "test-xi2",
+      "<device>",
+      test_xi2,
+    },
 #endif
     { "list-props",
       "<device> [<device> ...]",
@@ -131,29 +131,31 @@ static entry drivers[] =
       "<device> <property>",
       delete_prop
     },
-    {0, 0, 0
+    { "set-prop",
+      "<device> <property> <val> [<val> ...]",
+      set_prop
+    },
+    {NULL, NULL, NULL
     }
 };
 
-static Bool
-is_xinput_present(Display	*display)
+int
+xinput_version(Display	*display)
 {
     XExtensionVersion	*version;
-    Bool		present;
+    static int vers = -1;
+
+    if (vers != -1)
+        return vers;
 
-#if HAVE_XI2
-    version = XQueryInputVersion(display, XI_2_Major, XI_2_Minor);
-#else
     version = XGetExtensionVersion(display, INAME);
-#endif
 
     if (version && (version != (XExtensionVersion*) NoSuchExtension)) {
-	present = version->present;
+	vers = version->major_version;
 	XFree(version);
-	return present;
-    } else {
-	return False;
     }
+
+    return vers;
 }
 
 XDeviceInfo*
@@ -167,7 +169,7 @@ find_device_info(Display	*display,
     int		num_devices;
     int		len = strlen(name);
     Bool	is_id = True;
-    XID		id;
+    XID		id = (XID)-1;
 
     for(loop=0; loop<len; loop++) {
 	if (!isdigit(name[loop])) {
@@ -191,6 +193,7 @@ find_device_info(Display	*display,
 	                "Warning: There are multiple devices named \"%s\".\n"
 	                "To ensure the correct one is selected, please use "
 	                "the device ID instead.\n\n", name);
+		return NULL;
 	    } else {
 		found = &devices[loop];
 	    }
@@ -199,6 +202,41 @@ find_device_info(Display	*display,
     return found;
 }
 
+#ifdef HAVE_XI2
+XIDeviceInfo*
+xi2_find_device_info(Display *display, char *name)
+{
+    XIDeviceInfo *info;
+    int ndevices;
+    Bool is_id = True;
+    int i, id = -1;
+
+    for(i = 0; i < strlen(name); i++) {
+	if (!isdigit(name[i])) {
+	    is_id = False;
+	    break;
+	}
+    }
+
+    if (is_id) {
+	id = atoi(name);
+    }
+
+    info = XIQueryDevice(display, XIAllDevices, &ndevices);
+    for(i = 0; i < ndevices; i++)
+    {
+        if ((is_id && info[i].deviceid == id) ||
+                (!is_id && strcmp(info[i].name, name) == 0))
+        {
+            return &info[i];
+        }
+    }
+
+    XIFreeDeviceInfo(info);
+    return NULL;
+}
+#endif
+
 static void
 usage(void)
 {
@@ -235,7 +273,7 @@ main(int argc, char * argv[])
     func = argv[1];
     while((*func) == '-') func++;
 
-    if (!is_xinput_present(display)) {
+    if (!xinput_version(display)) {
 	fprintf(stderr, "%s extension not available\n", INAME);
 	return EXIT_FAILURE;
     }
@@ -245,6 +283,7 @@ main(int argc, char * argv[])
 	    int	r = (*driver->func)(display, argc-2, argv+2,
 				    driver->func_name, driver->arg_desc);
 	    XSync(display, False);
+	    XCloseDisplay(display);
 	    return r;
 	}
 	driver++;
